#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#else
double omp_get_wtime()
{return 0;}
#endif
//
//  Pour compiler
//  gcc-9 -fopenmp -o choleskiSeq choleskiSeq.c
//

// C = A^T A C est une matrice définie positive 
void gemm_ATA(double *A, double *C, int n) {
  for(int i=0; i<n; i++) {
		for(int j=0; j<n; j++) {
			double sum = 0;
			for(int k=0; k<n; k++) {
				sum += A[i*n+k]*A[j*n+k];
			}
			C[i*n+j] = sum;
		}
	}
}
void gemm_LLT(double *L, double *C, int n) {
  for(int i=0; i<n; i++) {
		for(int j=0; j<i+1; j++) {
			double sum = 0;
			for(int k=0; k<=j; k++) {
				sum += L[i*n+k]*L[j*n+k];
			}
			C[i*n+j] = sum;
			C[j*n+i] = sum;
		}
	}
}
////////////////////////////////////////////////////
//  version classique de l'algorithme de cholesky //
////////////////////////////////////////////////////

double *cholesky(double *A, int n) {
  double *L = (double*)calloc(n * n, sizeof(double));
  if (L == NULL)
    exit(EXIT_FAILURE);
 	
	for (int j = 0; j <n; j++) {
		
		double s = 0;
    for (int k = 0; k < j; k++) {
			s += L[j * n + k] * L[j * n + k];
		}
		L[j * n + j] = sqrt(A[j * n + j] - s);
		for (int i = j+1; i <n; i++) {
      double s = 0;
      for (int k = 0; k < j; k++) {
        s += L[i * n + k] * L[j * n + k];
			}
      L[i * n + j] = (1.0 / L[j * n + j] * (A[i * n + j] - s));
    }
	}
  return L;
}
	 
////////////////////////////////////////////////////


void show_matrix(double *A, int n) {
    for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++)
            printf("%2.5e ", A[i * n + j]);
        printf("\n");
    }
}
void write_matrix(double*mat, int n){
  FILE *fp=fopen("matrix.bin", "w");

  fwrite(&n, 1,sizeof(int),fp);
  fwrite(mat, n*n,sizeof(double),fp);
}
double* read_matrix(double *mat, int *n){
  FILE *fp = fopen("matrix.bin", "r");
  fread(n, 1,sizeof(int),fp);
    
  printf("size %d \n",*n);
  mat  = (double*)malloc(sizeof(double)*(*n)*(*n));
  fread(mat, (*n)*(*n),sizeof(double),fp);
  return mat;
}

double * build_pd_matrix(int n){

	double *tmp = (double*)malloc(sizeof(double)*n*n);
  // Construit une matrice symetrique aléatoire
	for(int i = 0 ; i < n ; i++) {
		for(int j = i ; j < n ; j++) {
			double element = 1.0*rand()/RAND_MAX;
			tmp[i*n+j] = element;
			tmp[j*n+i] = element;
		}
	}
	double *mat = (double*)malloc(sizeof(double)*n*n);
  //make a positive-definite matrix
	gemm_ATA(tmp, mat, n);
  //
  free(tmp);
  //
  return mat ;
}
int check_factors(double*L, double*M, int n){

	double *LTL = (double*)malloc(sizeof(double)*n*n);
  //make a positive-definite matrix
	gemm_LLT(L, LTL, n);
  double error = 0.0;

	for(int i = 0 ; i < n*n ; i++) {
    error = fmax(fabs(LTL[i] -M[i]),error);
    //        printf(" elt:  %d %f %f \n",i,LTL[i],aM[i]);	
  }

  printf("Error:  %f \n",error);	
  return (error < 1.e-14) ;

}
int main() {
	
	int n = 0;
	printf("Construction d'une matice PD \n");	

	double dtime;

	dtime = omp_get_wtime();
  double *mat = NULL;
  mat =read_matrix(mat,&n) ;
  //  mat = build_pd_matrix(n);
  dtime = omp_get_wtime() - 	dtime ;
  printf("Matrice construite en %f s\n",dtime);	
  printf("Matrice   %dx%d s\n",n,n);	

	printf("Factorisation de Cholesky \n");	
	dtime = omp_get_wtime();
	double *L = cholesky(mat, n); //warm up OpenMP
  dtime = omp_get_wtime() - 	dtime ;
  printf("Factorisation de Choleski en %f s\n",dtime);	
	printf("Check the factorisation \n");	
  int check = check_factors(L,mat,n) ;

  printf("error %d \n",check);	
  free(mat);
  free(L);


    return 0;
	
}

