#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#else
double omp_get_wtime()
{return 0;}
#endif
//
//  Pour compiler
//  gcc-9 -fopenmp -o generateMatrix generateMatrix.c
//  ./generateMatrix
// Changerla valeur de n (ligne 86)
// C = A^T A C est une matrice définie positive 
void gemm_ATA(double *A, double *C, int n) {
  for(int i=0; i<n; i++) {
		for(int j=0; j<n; j++) {
			double sum = 0;
			for(int k=0; k<n; k++) {
				sum += A[i*n+k]*A[j*n+k];
			}
			C[i*n+j] = sum;
		}
	}
}
void gemm_LLT(double *L, double *C, int n) {
  for(int i=0; i<n; i++) {
		for(int j=0; j<i+1; j++) {
			double sum = 0;
			for(int k=0; k<=j; k++) {
				sum += L[i*n+k]*L[j*n+k];
			}
			C[i*n+j] = sum;
			C[j*n+i] = sum;
		}
	}
}

void show_matrix(double *A, int n) {
    for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++)
            printf("%2.5e ", A[i * n + j]);
        printf("\n");
    }
}
void write_matrix(double*mat, int n){
  FILE *fp=fopen("matrix.bin", "w");

  fwrite(&n, 1,sizeof(int),fp);
  fwrite(mat, n*n,sizeof(double),fp);
}
double* read_matrix(double *mat, int *n){
    FILE *fp = fopen("matrix.bin", "r");
    fread(n, 1,sizeof(int),fp);
    
    printf("size %d \n",*n);
    mat  = (double*)malloc(sizeof(double)*(*n)*(*n));
    fread(mat, (*n)*(*n),sizeof(double),fp);
    show_matrix(mat, *n);
    return mat;
  }
double * build_pd_matrix(int n){

	double *tmp = (double*)malloc(sizeof(double)*n*n);
  // Construit une matrice symetrique aléatoire
	for(int i = 0 ; i < n ; i++) {
		for(int j = i ; j < n ; j++) {
			double element = 1.0*rand()/RAND_MAX;
			tmp[i*n+j] = element;
			tmp[j*n+i] = element;
		}
	}
	double *mat = (double*)malloc(sizeof(double)*n*n);
  //make a positive-definite matrix
	gemm_ATA(tmp, mat, n);
  //
  free(tmp);
  //
  return mat ;
}

int main() {
	
	int n = 100;
	printf("Construction d'une matice PD \n");	

	double dtime;

  double *mat = NULL;
  dtime = omp_get_wtime();
  mat = build_pd_matrix(n);
  dtime = omp_get_wtime() - 	dtime ;

  printf("Matrice construite en %f s\n",dtime);	
  write_matrix(mat, n);
  printf("Matrice ecrite dans matrix.bin  %d s\n",n);	

  free(mat);
    return 0;
	
}

